import math
from fractions import Fraction
from itertools import chain
from numbers import Rational
from typing import (Any,
                    Iterable,
                    Sequence,
                    Tuple,
                    TypeVar)

_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')

BASE = Fraction(10)


def ceil_half(value: int) -> int:
    return -(-value // 2)


def digits_count(value: int) -> int:
    return len('%i' % abs(value))


def identity(value: _T1) -> _T1:
    return value


try:
    lcm = math.lcm
except AttributeError:
    def lcm(left: int, right: int) -> int:
        left, right = abs(left), abs(right)
        return min(left, right) * (max(left, right) // math.gcd(left, right))


def positiveness_to_sign(flag: bool) -> int:
    return 2 * flag - 1


def rational_sqrt_lower_bound(value: Rational) -> Rational:
    return Fraction(sqrt_floor(value.numerator * value.denominator),
                    value.denominator)


def rational_sqrt_upper_bound(value: Rational) -> Rational:
    return Fraction(sqrt_ceil(value.numerator * value.denominator),
                    value.denominator)


def sqrt_ceil(value: int) -> int:
    value_sqrt_floor = sqrt_floor(value)
    return value_sqrt_floor + (value != square(value_sqrt_floor))


try:
    sqrt_floor = math.isqrt
except AttributeError:
    def sqrt_floor(value: int) -> int:
        if value > 0:
            candidate = 1 << ((value.bit_length() + 1) >> 1)
            while True:
                next_candidate = (candidate + value // candidate) >> 1
                if next_candidate >= candidate:
                    return candidate
                candidate = next_candidate
        else:
            assert not value, 'Argument should be non-negative.'
            return 0


def perfect_sqrt(value: int, alternative: int = 1) -> int:
    candidate = sqrt_floor(value)
    return candidate if square(candidate) == value else alternative


def square(value: Any) -> Any:
    return value * value


def to_binary_digits(value: int) -> Iterable[int]:
    for _ in range(value.bit_length()):
        yield value % 2
        value >>= 1


try:
    import _symba
except ImportError:
    def to_square_free(value: int) -> int:
        for candidate in _factors_candidates(value):
            factor_squared = candidate * candidate
            quotient, remainder = divmod(value, factor_squared)
            if not remainder:
                return to_square_free(quotient)
        return value


    def _factors_candidates(value: int) -> Iterable[int]:
        return chain((2,), range(3, sqrt_floor(value) + 1, 2))
else:
    to_square_free = _symba.to_square_free


def transpose(pairs_sequence: Sequence[Tuple[_T1, _T2]]
              ) -> Tuple[Sequence[_T1], Sequence[_T2]]:
    return tuple(zip(*pairs_sequence))



import math
from abc import abstractmethod
from fractions import Fraction
from numbers import (Rational,
                     Real)
from typing import (Any,
                    Optional,
                    Tuple,
                    Union)

from reprit.base import generate_repr

from .expression import Expression
from .hints import SqrtEvaluator
from .utils import (digits_count,
                    identity,
                    positiveness_to_sign,
                    perfect_sqrt,
                    square)


class Constant(Expression):
    @property
    def degree(self) -> int:
        return 0

    @property
    @abstractmethod
    def value(self) -> Real:
        """Returns value of the constant."""

    def evaluate(self, sqrt_evaluator: Optional[SqrtEvaluator] = None) -> Real:
        return self.value

    def is_positive(self) -> bool:
        return self.value > 0

    def lower_bound(self) -> Real:
        return self.value

    upper_bound = lower_bound

    def __eq__(self, other: Any) -> Any:
        return (self.value == other
                if isinstance(other, Real)
                else (isinstance(other, Constant)
                      and self.value == other.value
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __hash__(self) -> int:
        return hash(self.value)

    def __str__(self) -> str:
        return str(self.value)


class Finite(Constant):
    """Represents rational number."""
    is_finite = True

    __slots__ = '_value',

    def __init__(self, value: Real = 0) -> None:
        self._value = Fraction(value)

    @property
    def value(self) -> Rational:
        return self._value

    def evaluate(self, sqrt_evaluator: Optional[SqrtEvaluator] = None) -> Real:
        return self.value

    def extract_common_denominator(self) -> Tuple[int, 'Finite']:
        return self.value.denominator, Finite(self.value.numerator)

    def extract_common_numerator(self) -> Tuple[int, 'Finite']:
        return self.value.numerator, One / self.value.denominator

    def inverse(self) -> 'Finite':
        return Finite(Fraction(self.value.denominator, self.value.numerator))

    def is_positive(self) -> bool:
        return self.value > 0

    def perfect_sqrt(self) -> Expression:
        return Finite(Fraction(perfect_sqrt(self.value.numerator),
                               perfect_sqrt(self.value.denominator)))

    def significant_digits_count(self) -> int:
        return digits_count(self._value.limit_denominator(1).numerator)

    def square(self) -> 'Finite':
        return Finite(square(self.value))

    def __add__(self, other: Union[Real, 'Finite']) -> 'Finite':
        other = to_expression(other)
        return ((Finite(self.value + other.value)
                 if isinstance(other, Finite)
                 else other.__radd__(self))
                if isinstance(other, Expression)
                else NotImplemented)

    def __bool__(self) -> bool:
        return bool(self.value)

    def __mul__(self, other: Union[Real, 'Finite']) -> 'Finite':
        other = to_expression(other)
        return ((Finite(self.value * other.value)
                 if isinstance(other, Finite)
                 else other.__rmul__(self))
                if isinstance(other, Expression)
                else NotImplemented)

    def __neg__(self) -> 'Finite':
        return Finite(-self.value)

    def __radd__(self, other: Union[Real, 'Finite']) -> 'Finite':
        return (to_expression(other) + self
                if isinstance(other, Real)
                else NotImplemented)

    __repr__ = generate_repr(__init__)

    def __rmul__(self, other: Union[Real, 'Finite']) -> 'Finite':
        return (to_expression(other) * self
                if isinstance(other, Real)
                else NotImplemented)


Zero, One = Finite(0), Finite(1)


class Infinite(Constant):
    is_finite = False

    @property
    def degree(self) -> int:
        return 0

    @property
    def value(self) -> Real:
        return positiveness_to_sign(self.is_positive()) * math.inf

    __slots__ = '_is_positive',

    def __init__(self, is_positive: bool) -> None:
        self._is_positive = is_positive

    def evaluate(self, sqrt_evaluator: Optional[SqrtEvaluator] = None) -> Real:
        return self.value

    def extract_common_denominator(self) -> Tuple[int, 'Expression']:
        return 1, self

    def extract_common_numerator(self) -> Tuple[int, 'Expression']:
        return 1, self

    def inverse(self) -> 'Expression':
        return Zero

    def is_positive(self) -> bool:
        return self._is_positive

    perfect_sqrt = identity

    def significant_digits_count(self) -> int:
        return 0

    def square(self) -> 'Expression':
        return Infinity

    def __add__(self, other: Union[Real, 'Expression']) -> Constant:
        other = to_expression(other)
        return ((self
                 if (other.is_finite
                     or (other is not NaN
                         and self.is_positive() is other.is_positive()))
                 else NaN)
                if isinstance(other, Expression)
                else NotImplemented)

    def __ge__(self, other: Union[Real, 'Expression']) -> bool:
        other = to_expression(other)
        return (other is not NaN and (self.is_positive() or self == other)
                if isinstance(other, (Real, Expression))
                else NotImplemented)

    def __gt__(self, other: Union[Real, 'Expression']) -> bool:
        other = to_expression(other)
        return (other is not NaN and self.is_positive() and self != other
                if isinstance(other, Expression)
                else NotImplemented)

    def __le__(self, other: Union[Real, 'Expression']) -> bool:
        other = to_expression(other)
        return (other is not NaN and (not self.is_positive() or self == other)
                if isinstance(other, Expression)
                else NotImplemented)

    def __lt__(self, other: Union[Real, 'Expression']) -> bool:
        other = to_expression(other)
        return (other is not NaN and not self.is_positive() and self != other
                if isinstance(other, Expression)
                else NotImplemented)

    def __mul__(self, other: Union[Real, 'Expression']) -> Constant:
        other = to_expression(other)
        return (((Infinity
                  if self.is_positive() is other.is_positive()
                  else -Infinity)
                 if other and other is not NaN
                 else NaN)
                if isinstance(other, Expression)
                else NotImplemented)

    def __neg__(self) -> 'Expression':
        return Infinite(not self.is_positive())

    __radd__ = __add__
    __repr__ = generate_repr(__init__)
    __rmul__ = __mul__


Infinity = Infinite(True)


class _NaN(Constant):
    is_finite = False
    value = math.nan
    _instance = None

    def __new__(cls) -> '_NaN':
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    __slots__ = ()

    def extract_common_denominator(self) -> Tuple[int, 'Expression']:
        return 1, self

    def extract_common_numerator(self) -> Tuple[int, 'Expression']:
        return 1, self

    def inverse(self) -> 'Expression':
        return self

    def is_positive(self) -> bool:
        return False

    perfect_sqrt = identity

    def significant_digits_count(self) -> int:
        return 0

    square = identity

    def __add__(self, other: Union[Real, 'Expression']) -> 'Expression':
        return self

    def __ge__(self, other: Union[Real, 'Expression']) -> bool:
        return (False
                if isinstance(other, (Real, Expression))
                else NotImplemented)

    def __gt__(self, other: Union[Real, 'Expression']) -> bool:
        return (False
                if isinstance(other, (Real, Expression))
                else NotImplemented)

    def __le__(self, other: Union[Real, 'Expression']) -> bool:
        return (False
                if isinstance(other, (Real, Expression))
                else NotImplemented)

    def __lt__(self, other: Union[Real, 'Expression']) -> bool:
        return (False
                if isinstance(other, (Real, Expression))
                else NotImplemented)

    def __mul__(self, other: Union[Real, 'Expression']) -> 'Expression':
        return self

    __neg__ = identity

    def __radd__(self, other: Union[Real, 'Expression']) -> 'Expression':
        return self

    def __repr__(self) -> str:
        return 'NaN'

    def __rmul__(self, other: Union[Real, 'Expression']) -> 'Expression':
        return self


NaN = _NaN()


def to_expression(other: Union[Real, Expression]) -> Expression:
    return ((Finite(other)
             if isinstance(other, Rational) or math.isfinite(other)
             else (Infinite(other > 0)
                   if math.isinf(other)
                   else NaN))
            if isinstance(other, Real)
            else other)



import platform
from pathlib import Path

from setuptools import (find_packages,
                        setup)

import symba

project_base_url = 'https://github.com/lycantropos/symba/'


def read_file(path_string: str) -> str:
    return Path(path_string).read_text(encoding='utf-8')


parameters = dict(
        name=symba.__name__,
        packages=find_packages(exclude=('tests', 'tests.*')),
        version=symba.__version__,
        description=symba.__doc__,
        long_description=read_file('README.md'),
        long_description_content_type='text/markdown',
        author='Azat Ibrakov',
        author_email='azatibrakov@gmail.com',
        license='MIT License',
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Python :: Implementation :: PyPy',
        ],
        url=project_base_url,
        download_url=project_base_url + 'archive/master.zip',
        python_requires='>=3.5',
        setup_requires=read_file('requirements-setup.txt'),
        install_requires=read_file('requirements.txt'))
if platform.python_implementation() == 'CPython':
    import sys
    import tempfile
    from collections import defaultdict
    from datetime import date
    from distutils.ccompiler import CCompiler
    from distutils.errors import CompileError
    from glob import glob
    from setuptools import Extension
    from setuptools.command.build_ext import build_ext


    def has_flag(compiler: CCompiler, name: str) -> bool:
        """
        Detects whether a flag name is supported on the specified compiler.
        """
        with tempfile.NamedTemporaryFile('w',
                                         suffix='.cpp') as file:
            file.write('int main (int argc, char **argv) { return 0; }')
            try:
                compiler.compile([file.name],
                                 extra_postargs=[name])
            except CompileError:
                return False
        return True


    def cpp_flag(compiler: CCompiler) -> str:
        """
        Returns the -std=c++[11|...] compiler flag.
        The newer version is preferred when available.
        """
        flags = ['-std=c++{}'.format(str(year)[2:])
                 for year in range(2011, date.today().year + 1, 3)]
        for flag in reversed(flags):
            if has_flag(compiler, flag):
                return flag
        raise RuntimeError('Unsupported compiler: '
                           'at least C++11 support is needed.')


    class BuildExt(build_ext):
        """A custom build extension for adding compiler-specific options."""
        compile_args = defaultdict(list,
                                   {'msvc': ['/EHsc'],
                                    'unix': []})
        link_args = defaultdict(list,
                                {'msvc': [],
                                 'unix': []})

        if sys.platform == 'darwin':
            darwin_args = ['-stdlib=libc++', '-mmacosx-version-min=10.7']
            compile_args['unix'] += darwin_args
            link_args['unix'] += darwin_args

        def build_extensions(self) -> None:
            compiler_type = self.compiler.compiler_type
            compile_args = self.compile_args[compiler_type]
            link_args = self.link_args[compiler_type]
            if compiler_type == 'unix':
                compile_args.append(cpp_flag(self.compiler))
                if has_flag(self.compiler, '-fvisibility=hidden'):
                    compile_args.append('-fvisibility=hidden')
            define_macros = [('VERSION_INFO', self.distribution.get_version())]
            for extension in self.extensions:
                extension.extra_compile_args = compile_args
                extension.extra_link_args = link_args
                extension.define_macros = define_macros
            super().build_extensions()


    class LazyPybindInclude:
        def __str__(self) -> str:
            import pybind11
            return pybind11.get_include()


    parameters.update(
            cmdclass={'build_ext': BuildExt},
            ext_modules=[Extension('_' + symba.__name__,
                                   glob('src/*.cpp'),
                                   include_dirs=[LazyPybindInclude()],
                                   language='c++')],
            zip_safe=False)
setup(**parameters)



import math
from numbers import Real
from typing import (TYPE_CHECKING,
                    Any,
                    Optional,
                    Tuple,
                    Union)

from reprit.base import generate_repr

from symba.core.utils import (ceil_half,
                              rational_sqrt_lower_bound,
                              rational_sqrt_upper_bound)
from . import context
from .constant import (Constant,
                       Finite,
                       NaN,
                       One,
                       Zero,
                       to_expression)
from .expression import Expression
from .hints import SqrtEvaluator

if TYPE_CHECKING:
    from .form import Form


class Term(Expression):
    """Represents square root of the expression."""
    is_finite = True

    @classmethod
    def from_components(cls,
                        scale: Finite,
                        argument: Expression) -> Expression:
        if not (scale and argument):
            return Zero
        denominator, argument = argument.extract_common_denominator()
        scale /= denominator
        argument *= denominator
        argument_perfect_sqrt = argument.perfect_sqrt()
        argument_perfect_part = argument_perfect_sqrt.square()
        return (argument_perfect_sqrt * scale
                if argument == argument_perfect_part
                else (argument_perfect_sqrt
                      * cls(scale, argument / argument_perfect_part)))

    __slots__ = 'argument', 'scale'

    def __init__(self, scale: Finite, argument: Expression) -> None:
        self.scale, self.argument = scale, argument

    @property
    def degree(self) -> int:
        return self.argument.degree + 1

    def evaluate(self, sqrt_evaluator: Optional[SqrtEvaluator] = None) -> Real:
        return (self.scale.evaluate(sqrt_evaluator)
                * ((context.sqrt_evaluator.get()
                    if sqrt_evaluator is None
                    else sqrt_evaluator)
                   (self.argument.evaluate(sqrt_evaluator))))

    def extract_common_denominator(self) -> Tuple[int, 'Term']:
        common_denominator, scale = self.scale.extract_common_denominator()
        return common_denominator, Term(scale, self.argument)

    def extract_common_numerator(self) -> Tuple[int, 'Term']:
        common_numerator, scale = self.scale.extract_common_numerator()
        return common_numerator, Term(scale, self.argument)

    def inverse(self) -> 'Term':
        scale = self.scale.inverse()
        denominator, argument = (self.argument.inverse()
                                 .extract_common_denominator())
        scale /= denominator
        argument *= denominator
        return Term(scale, argument)

    def is_positive(self) -> bool:
        return self.scale.is_positive()

    def lower_bound(self) -> Real:
        return (rational_sqrt_lower_bound(self.square().lower_bound())
                if self.is_positive()
                else -(-self).upper_bound())

    def perfect_sqrt(self) -> Expression:
        return self.scale.perfect_sqrt()

    def significant_digits_count(self) -> int:
        return ceil_half(self.square().significant_digits_count())

    def square(self) -> Expression:
        return self.scale.square() * self.argument

    def upper_bound(self) -> Real:
        return (rational_sqrt_upper_bound(self.square().upper_bound())
                if self.is_positive()
                else -(-self).lower_bound())

    def __add__(self, other: Union[Real, Expression]) -> Expression:
        from .form import Form
        return (self._add_constant(other)
                if isinstance(other, (Real, Constant))
                else (Form.from_components([self, other])
                      if isinstance(other, Term)
                      else NotImplemented))

    def __eq__(self, other: Any) -> Any:
        return (isinstance(other, Term)
                and self.is_positive() is other.is_positive()
                and self.square() == other.square()
                if isinstance(other, Expression)
                else NotImplemented)

    def __ge__(self, other: Union[Real, Expression]) -> bool:
        other = to_expression(other)
        return (((not other.is_positive()
                  or other.square() <= self.square()
                  if self.is_positive()
                  else (not other.is_positive()
                        and other.square() >= self.square()))
                 if other.is_finite
                 else other <= self)
                if isinstance(other, Expression)
                else NotImplemented)

    def __gt__(self, other: Union[Real, Expression]) -> bool:
        other = to_expression(other)
        return (((not other.is_positive() or other.square() < self.square()
                  if self.is_positive()
                  else (not other.is_positive()
                        and other.square() > self.square()))
                 if other.is_finite
                 else other < self)
                if isinstance(other, Expression)
                else NotImplemented)

    def __hash__(self) -> int:
        return hash((self.is_positive(), self.square()))

    def __le__(self, other: Union[Real, Expression]) -> bool:
        other = to_expression(other)
        return (((other.is_positive() and other.square() >= self.square()
                  if self.is_positive()
                  else (other.is_positive()
                        or other.square() <= self.square()))
                 if other.is_finite
                 else other >= self)
                if isinstance(other, Expression)
                else NotImplemented)

    def __lt__(self, other: Union[Real, Expression]) -> bool:
        other = to_expression(other)
        return (((other.is_positive() and other.square() > self.square()
                  if self.is_positive()
                  else (other.is_positive()
                        or other.square() < self.square()))
                 if other.is_finite
                 else other > self)
                if isinstance(other, Expression)
                else NotImplemented)

    def __mul__(self, other: Union[Real, Expression]) -> Expression:
        return (self._multiply_by_constant(other)
                if isinstance(other, (Real, Constant))
                else (self._multiply_by_term(other)
                      if isinstance(other, Term)
                      else NotImplemented))

    def __neg__(self) -> 'Term':
        return Term(-self.scale, self.argument)

    def __radd__(self, other: Union[Real, Expression]) -> Expression:
        return (self._add_constant(other)
                if isinstance(other, (Real, Constant))
                else NotImplemented)

    __repr__ = generate_repr(__init__)

    def __rmul__(self, other: Union[Real, Expression]) -> Expression:
        return (self._multiply_by_constant(other)
                if isinstance(other, (Real, Constant))
                else NotImplemented)

    def __str__(self) -> str:
        return ((''
                 if self.scale == One
                 else ('-'
                       if self.scale == -One
                       else '{} * '.format(self.scale)))
                + 'sqrt({})'.format(self.argument))

    def _add_constant(self, other: Union[Real, Constant]) -> Expression:
        from .form import Form
        other = to_expression(other)
        return ((Form([self], other) if other else self)
                if other.is_finite
                else other)

    def _multiply_by_constant(self,
                              other: Union[Real, Constant]) -> Expression:
        other = to_expression(other)
        return ((Term(self.scale * other, self.argument) if other else other)
                if other.is_finite
                else (other if other is NaN or self.is_positive() else -other))

    def _multiply_by_term(self, other: 'Term') -> Expression:
        scale = self.scale * other.scale
        if self.argument == other.argument:
            return scale * self.argument
        argument, other_argument = ((self.argument, other.argument)
                                    if self.argument < other.argument
                                    else (other.argument, self.argument))
        if not other_argument % argument:
            return argument * Term.from_components(scale,
                                                   other_argument / argument)
        elif (isinstance(argument, Finite)
              and isinstance(other_argument, Finite)):
            arguments_gcd = math.gcd(argument.value.numerator,
                                     other_argument.value.numerator)
            argument /= arguments_gcd
            other_argument /= arguments_gcd
            scale *= arguments_gcd
        return Term.from_components(scale, argument * other_argument)
